import dataclasses
from typing import Any

from theia.core import primitives

_TEST_BBOX_DICT = {"x0": 10, "y0": 20, "x1": 100, "y1": 200}
_TEST_DETECTION_DICT = {"bounding_box": _TEST_BBOX_DICT, "confidence": 0.85}
_TEST_EVALUATED_DETECTION_DICT = {
    "bounding_box": _TEST_BBOX_DICT,
    "confidence": 0.85,
    "is_close": True,
    "is_confident": True,
}
_TEST_EVALUATED_INFERENCE_DICT = {
    "pts": 12345,
    "evaluated_detections": [
        _TEST_EVALUATED_DETECTION_DICT,
        _TEST_EVALUATED_DETECTION_DICT,
    ],
}
_TEST_STREAM_STATUS_DICT = {
    "input_fps": 30.0,
    "latest_evaluated_inference": _TEST_EVALUATED_INFERENCE_DICT,
    "inference_fps": 25.0,
    "safety_state": 1,  # SafetyState.HEALTHY
}
_TEST_SYSTEM_STATUS_DICT = {
    "stream_statuses": {
        "cam1": _TEST_STREAM_STATUS_DICT,
        "cam2": _TEST_STREAM_STATUS_DICT,
    }
}


def _validate_evaluated_detection(evaluated_detection: Any) -> None:
    assert isinstance(evaluated_detection.bounding_box, primitives.BoundingBox)


def test_evaluated_detection_from_dict() -> None:
    """Tests that EvaluatedDetection.from_dict properly reconstructs nested objects."""
    evaluated_detection = primitives.EvaluatedDetection.from_dict(
        _TEST_EVALUATED_DETECTION_DICT
    )
    _validate_evaluated_detection(evaluated_detection)


def _validate_evaluated_inference(evaluated_inference: Any) -> None:
    for evaluated_detection in evaluated_inference.evaluated_detections:
        _validate_evaluated_detection(evaluated_detection)


def test_evaluated_inference_from_dict() -> None:
    """Tests that EvaluatedInference.from_dict properly reconstructs nested lists."""
    evaluated_inference = primitives.EvaluatedInference.from_dict(
        _TEST_EVALUATED_INFERENCE_DICT
    )
    _validate_evaluated_inference(evaluated_inference)


def _validate_stream_status(stream_status: Any) -> None:
    assert isinstance(stream_status.safety_state, primitives.SafetyState)
    assert isinstance(
        stream_status.latest_evaluated_inference, primitives.EvaluatedInference
    )


def test_stream_status_from_dict() -> None:
    """Tests that StreamStatus.from_dict properly handles enums and nested objects."""
    status = primitives.StreamStatus.from_dict(_TEST_STREAM_STATUS_DICT)
    _validate_stream_status(status)


def _validate_system_status(system_status: Any) -> None:
    for stream_status in system_status.stream_statuses.values():
        _validate_stream_status(stream_status)


def test_system_status_from_dict() -> None:
    """Tests that SystemStatus.from_dict properly reconstructs nested StreamStatus objects."""
    sys_status = primitives.SystemStatus.from_dict(_TEST_SYSTEM_STATUS_DICT)
    _validate_system_status(sys_status)


def test_stream_status_from_dict_idempotent() -> None:
    """Tests that StreamStatus.from_dict can handle already-converted objects."""
    status = primitives.StreamStatus.from_dict(_TEST_STREAM_STATUS_DICT)
    status_dict = dataclasses.asdict(status)
    loaded_status = primitives.StreamStatus.from_dict(status_dict)
    assert status == loaded_status



from theia.client import client
from theia.core import config, primitives

BoundingBox = primitives.BoundingBox
CameraConfig = config.CameraConfig
Client = client.Client
Detection = primitives.Detection
EvaluatedDetection = primitives.EvaluatedDetection
EvaluatedInference = primitives.EvaluatedInference
SafetyState = primitives.SafetyState
ServerConfig = config.ServerConfig
StreamStatus = primitives.StreamStatus
SystemStatus = primitives.SystemStatus



def status_path() -> str:
    """Returns the path component for the status endpoint."""
    return "/api/status"


def ready_path() -> str:
    """Returns the path component for the ready endpoint."""
    return "/api/ready"


def _stream_image_path_prefix(stream_id: str) -> str:
    """Returns the path prefix component for a stream's image endpoint."""
    return f"/api/{stream_id}/image"


def stream_image_path(stream_id: str, pts: int) -> str:
    """Returns the path component for a stream's image endpoint."""
    return f"{_stream_image_path_prefix(stream_id)}/{pts}"


def stream_image_path_template(stream_id: str) -> str:
    """Returns the path template for a stream's image endpoint."""
    return _stream_image_path_prefix(stream_id) + "/{pts:int}"



from __future__ import annotations

import dataclasses
import enum
from typing import Any


@dataclasses.dataclass(kw_only=True)
class SafetyConfig:
    """Safety parameters used in determining whether a pedestrian has been detected."""

    close_bbox_diagonal_px: float
    decision_threshold: float
    min_throughput_hz: float
    monitor_window_s: float


@dataclasses.dataclass(frozen=True, kw_only=True)
class BoundingBox:
    """Represents a 2D bounding on the image plane defined by its top left and bottom
    right corners. The coordinates are expressed in pixels with the X coordinates
    measured from the image's left edge, and the Y coordinates are measured from the top
    of the image.

    p0 -------
    | |__O__| |
    |    |    |
    |  _/|_   |
     ------- p1

    Where: p0 = (x0, y0), p1 = (x1, y1)
    """

    x0: int
    y0: int
    x1: int
    y1: int

    @classmethod
    def from_xywh(cls, x: int, y: int, width: int, height: int) -> BoundingBox:
        return cls(x0=x, y0=y, x1=x + width, y1=y + height)

    @property
    def width(self) -> int:
        return self.x1 - self.x0

    @property
    def height(self) -> int:
        return self.y1 - self.y0

    @property
    def diagonal(self) -> float:
        """Length of the diagonal of this bounding box in pixels."""
        return (self.width**2 + self.height**2) ** 0.5

    @property
    def area(self) -> int:
        return self.width * self.height

    def intersection(self, other: BoundingBox) -> int:
        intersect_w = min(self.x1, other.x1) - max(self.x0, other.x0)
        intersect_h = min(self.y1, other.y1) - max(self.y0, other.y0)
        if intersect_w < 0 or intersect_h < 0:
            return 0
        else:
            return intersect_w * intersect_h

    def union(self, other: BoundingBox) -> int:
        return self.area + other.area - self.intersection(other)

    def iou(self, other: BoundingBox) -> float:
        """Calculates the Intersection over Union (IoU) of two bounding boxes.
        See https://en.wikipedia.org/wiki/Jaccard_index.
        """
        return self.intersection(other) / self.union(other)

    def to_xywh(self) -> tuple[int, int, int, int]:
        return self.x0, self.y0, self.width, self.height


def _parse_bounding_box(obj: Any) -> BoundingBox:
    if isinstance(obj, BoundingBox):
        return obj
    if isinstance(obj, dict):
        return BoundingBox(**obj)
    raise ValueError(f"Invalid type: {type(obj)}")


@dataclasses.dataclass(frozen=True, kw_only=True)
class Detection:
    """A detected object within an image."""

    bounding_box: BoundingBox
    confidence: float


@dataclasses.dataclass(frozen=True, kw_only=True)
class EvaluatedDetection(Detection):
    """A detection evaluated on whether the object is close and/or confident.

    Primarily exists for use in StreamStatus in WebSocket messages from Theia to
    clients.
    """

    is_close: bool
    is_confident: bool

    @classmethod
    def from_dict(cls, det_dict: dict) -> EvaluatedDetection:
        bounding_box = _parse_bounding_box(det_dict["bounding_box"])
        return cls(
            bounding_box=bounding_box,
            confidence=det_dict["confidence"],
            is_close=det_dict["is_close"],
            is_confident=det_dict["is_confident"],
        )


def _parse_evaluated_detection(obj: Any) -> EvaluatedDetection:
    if isinstance(obj, EvaluatedDetection):
        return obj
    if isinstance(obj, dict):
        return EvaluatedDetection.from_dict(obj)
    raise ValueError(f"Invalid type: {type(obj)}")


@dataclasses.dataclass(frozen=True, kw_only=True)
class EvaluatedInference:
    """An inference with EvaluatedDetections."""

    pts: int
    evaluated_detections: list[EvaluatedDetection]

    @classmethod
    def from_dict(cls, evaluated_inference_dict: dict) -> EvaluatedInference:
        evaluated_detections = [
            _parse_evaluated_detection(obj)
            for obj in evaluated_inference_dict["evaluated_detections"]
        ]
        pts = evaluated_inference_dict["pts"]
        return cls(pts=pts, evaluated_detections=evaluated_detections)


class SafetyState(enum.IntEnum):
    """A state that reflects safety conditions in a stream."""

    HEALTHY = enum.auto()
    PEDESTRIAN_PRESENT = enum.auto()
    LOW_INFERENCE_THROUGHPUT = enum.auto()
    LOW_INPUT_THROUGHPUT = enum.auto()


def _parse_latest_evaluated_inference(obj: Any) -> EvaluatedInference | None:
    if obj is None:
        return None
    if isinstance(obj, EvaluatedInference):
        return obj
    if isinstance(obj, dict):
        return EvaluatedInference.from_dict(obj)
    raise ValueError(f"Invalid type: {type(obj)}")


def _parse_safety_state(obj: Any) -> SafetyState:
    if isinstance(obj, SafetyState):
        return obj
    if isinstance(obj, int):
        return SafetyState(obj)
    raise ValueError(f"Invalid type: {type(obj)}")


@dataclasses.dataclass(kw_only=True)
class StreamStatus:
    """Comprises the current status of a stream.

    The input_fps reflects how fast frames are being received from the RTSP stream.
    The inference_fps reflects how fast inferences are being generated.
    The latest_evaluated_inference contains the most recent inference with
    safety-evaluated detections, or None if no inference has been processed yet.
    """

    input_fps: float
    inference_fps: float
    latest_evaluated_inference: EvaluatedInference | None
    safety_state: SafetyState

    @classmethod
    def from_dict(cls, status_dict: dict) -> StreamStatus:
        latest_evaluated_inference_obj = status_dict.get("latest_evaluated_inference")
        latest_evaluated_inference = _parse_latest_evaluated_inference(
            latest_evaluated_inference_obj
        )

        safety_state_obj = status_dict.get("safety_state")
        safety_state = _parse_safety_state(safety_state_obj)

        return cls(
            input_fps=status_dict["input_fps"],
            latest_evaluated_inference=latest_evaluated_inference,
            inference_fps=status_dict["inference_fps"],
            safety_state=safety_state,
        )


@dataclasses.dataclass(kw_only=True)
class SystemStatus:
    """Aggregates individual stream statuses with an overall system stafety state."""

    stream_statuses: dict[str, StreamStatus]

    @classmethod
    def from_dict(cls, status_dict: dict) -> SystemStatus:
        stream_status_dicts = status_dict["stream_statuses"]
        stream_statuses = {
            stream_id: StreamStatus.from_dict(stream_status_dict)
            for stream_id, stream_status_dict in stream_status_dicts.items()
        }
        return cls(stream_statuses=stream_statuses)



import dataclasses
import pathlib

import yaml


@dataclasses.dataclass(kw_only=True)
class TransformConfig:
    destack: bool | None = None
    # Crop region is given in the format (x0, y0, width, height) where (x0, y0) is the
    # top-left corner.
    crop: tuple[int, int, int, int] | None = None


@dataclasses.dataclass(kw_only=True)
class CameraConfig:
    resolution: tuple[int, int]
    rtsp_url: str
    transform: TransformConfig | None = None


@dataclasses.dataclass(kw_only=True)
class ServerConfig:
    cameras: dict[str, CameraConfig]
    close_bbox_diagonal_px: float
    http_port: int
    recording_lifetime_s: float
    status_frequency_hz: float

    @property
    def camera_ids(self) -> list[str]:
        return list(self.cameras.keys())


def _load_transform_config(transform_dict: dict) -> TransformConfig:
    """Load transform configuration from a dictionary."""
    crop = transform_dict.get("crop")
    # YAML doesn't have built in tuple support so we convert it here.
    if crop is not None:
        crop = tuple(crop)

    transform = TransformConfig(destack=transform_dict.get("destack"), crop=crop)
    return transform


def _load_camera_configs(cameras_dict: dict) -> dict[str, CameraConfig]:
    """Load camera configurations from a dictionary."""
    cameras = {}
    for cam_id, cam_dict in cameras_dict.items():
        if transform_dict := cam_dict.get("transform"):
            transform = _load_transform_config(transform_dict)
        else:
            transform = None

        cameras[cam_id] = CameraConfig(
            # YAML doesn't have built in tuple support so we convert it here.
            resolution=tuple(cam_dict["resolution"]),
            rtsp_url=cam_dict["rtsp_url"],
            transform=transform,
        )
    return cameras


def load(path: pathlib.Path) -> ServerConfig:
    """Loads the server configuration from a YAML file."""
    with open(path, "r") as f:
        loaded_cfg = yaml.safe_load(f)

    cameras = _load_camera_configs(loaded_cfg.get("cameras", {}))

    server_cfg = ServerConfig(
        cameras=cameras,
        close_bbox_diagonal_px=loaded_cfg["close_bbox_diagonal_px"],
        http_port=loaded_cfg["http_port"],
        recording_lifetime_s=loaded_cfg["recording_lifetime_s"],
        status_frequency_hz=loaded_cfg["status_frequency_hz"],
    )
    if len(server_cfg.cameras) == 0:
        raise ValueError("Configuration must specify at least one camera.")
    return server_cfg


def save(config: ServerConfig, path: pathlib.Path) -> None:
    """Saves the server configuration to a YAML file."""
    data = dataclasses.asdict(config)

    # YAML doesn't have built in tuple support so we convert it here.
    for cam_cfg in data["cameras"].values():
        cam_cfg["resolution"] = list(cam_cfg["resolution"])

    with open(path, "w") as f:
        yaml.safe_dump(data, f)



"""Test mocking utilities for core functionality."""

import contextlib
from typing import Generator
from unittest import mock


def _mock_time_time(time_s: float) -> mock._patch:
    """Monkeypatches time.time() to return the given time.

    Should be used as a context manager.
    """
    return mock.patch("time.time", mock.MagicMock(return_value=time_s))


def _mock_time_perf_counter(time_s: float) -> mock._patch:
    """Monkeypatches time.perf_counter() to return the given time. Since
    time.perf_counter() is used inside Timer, this mocks the time measured by Timer.

    Should be used as a context manager.
    """
    return mock.patch("time.perf_counter", mock.MagicMock(return_value=time_s))


@contextlib.contextmanager
def mock_time(time_s: float) -> Generator[None, None, None]:
    """Mocks the flow of time reported by time.perf_counter() and time.time()."""
    with _mock_time_perf_counter(time_s), _mock_time_time(time_s):
        yield
